
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <malloc.h>
#include <errno.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <chkstack.h>
#include <libsmbclient.h>

#include "../smbremote/smbc_api.h"
int smbc_redir_thread(void *pv);

pthread_key_t g_chkstack_key = -1;

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#if 0
extern void *_start, *_fini;
unsigned long checksum_smbclient(void)
{
    unsigned long v = 0;
    unsigned long *p;
    static unsigned long csum = 0;

    p = &_start;
    while (p < &_fini) {
	v += *p;
	p++;
    }
    if (!csum) {
	csum = v;
	printf("smbclient checksum is %lx\n", csum);
    }
    else if (v != csum) {
	printf("smbclient checksum mismatch %lx\n", v);
	sleep(1000);
    }
    return v;
}
#endif

void smbc_auth_fn(const char *server, const char *share,
		  char *workgroup, int wgmaxlen, char *username, int unmaxlen,
		  char *password, int pwmaxlen)
{   
    username[0] = password[0] = 0;
}

int smbc_dispatch_msg(SMBC_MSG *msg, SMBC_ACK_INFO *ack_info)
{
    int size;
    SMBC_OPEN_INFO *open_info;
    SMBC_READ_INFO *read_info;
    SMBC_WRITE_INFO *write_info;
    SMBC_LSEEK_INFO *lseek_info;
    SMBC_CLOSE_INFO *close_info;
    SMBC_UNLINK_INFO *unlink_info;
    SMBC_LOCK_INFO *lock_info;	
    SMBC_UNLOCK_INFO *unlock_info;
    SMBC_RENAME_INFO *rename_info;
    SMBC_OPENDIR_INFO *opendir_info;
    SMBC_CLOSEDIR_INFO *closedir_info;
    SMBC_READDIR_INFO *readdir_info;
    SMBC_TELLDIR_INFO *telldir_info;	
    SMBC_LSEEKDIR_INFO *lseekdir_info;
    SMBC_MKDIR_INFO *mkdir_info;
    SMBC_RMDIR_INFO *rmdir_info;
    SMBC_STAT_INFO *stat_info;
    SMBC_FSTAT_INFO *fstat_info;
    SMBC_CHOWN_INFO *chown_info;
    SMBC_CHMOD_INFO *chmod_info;
    SMBC_REMOVEUNUSEDSERVERS_INFO *rus_info;
    NMBC_NAMEQUERY_INFO *nq_info;

    CHKSTACK();
	errno=0;
    switch (msg->type) {
    case SMBC_MSG_OPEN:
	open_info = msg;
	//printf("smbc_message_loop: OPEN %s\n", open_info->fname);
	ack_info->value.iValue = smbc_open(open_info->fname, open_info->flags, open_info->mode);
	size = sizeof(int);
	break;
    case SMBC_MSG_READ:
	read_info = msg;
	//printf("smbc_message_loop: READ %d\n", read_info->count);
	ack_info->value.iValue = smbc_read(read_info->fd, read_info->buf, read_info->count);
	//printf("read %d bytes\n", ack_info->value.iValue);
	size = sizeof(int);
	break;
    case SMBC_MSG_WRITE:
	write_info = msg;
	//printf("smbc_message_loop: WRITE %d\n", write_info->count);
	ack_info->value.iValue = smbc_write(write_info->fd, write_info->buf, write_info->count);
	//printf("wrote %d bytes\n", ack_info->value.iValue);
	size = sizeof(int);
	break;
    case SMBC_MSG_LSEEK:
	lseek_info = msg;
	//printf("smbc_message_loop: LSEEK %d\n", lseek_info->fd);
	ack_info->value.llValue = smbc_lseek(lseek_info->fd, lseek_info->offset, lseek_info->whence);
	size = sizeof(long long);
	break;
    case SMBC_MSG_CLOSE:
	close_info = msg;
	//printf("smbc_message_loop: CLOSE %d\n", close_info->fd);
	ack_info->value.iValue = smbc_close(close_info->fd);
	size = sizeof(int);
	break;
    case SMBC_MSG_UNLINK:
	unlink_info = msg;
	//printf("smbc_message_loop: UNLINK %s\n", unlink_info->furl);
	ack_info->value.iValue = smbc_unlink(unlink_info->furl);
	size = sizeof(int);
	break;
    case SMBC_MSG_LOCK:
	lock_info = msg;
	//printf("smbc_message_loop: LOCK %d\n", lock_info->fd);
	ack_info->value.iValue = smbc_lock(lock_info->fd, lock_info->offset, lock_info->len, lock_info->timeout, lock_info->lock_type);
	size = sizeof(int);
	break;
    case SMBC_MSG_UNLOCK:
	unlock_info = msg;
	//printf("smbc_message_loop: UNLOCK %d\n", unlock_info->fd);
	ack_info->value.iValue = smbc_unlock(unlock_info->fd, unlock_info->offset, unlock_info->len);
	size = sizeof(int);
	break;
    case SMBC_MSG_RENAME:
	rename_info = msg;
	//printf("smbc_message_loop: RENAME %s -> %s\n", rename_info->ourl, rename_info->nurl);
	ack_info->value.iValue = smbc_rename(rename_info->ourl, rename_info->nurl);
	size = sizeof(int);
	break;
    case SMBC_MSG_OPENDIR:
	opendir_info = msg;
	//printf("smbc_message_loop: OPENDIR %s\n", opendir_info->fname);
	ack_info->value.iValue = smbc_opendir(opendir_info->fname);
	size = sizeof(int);
	break;
    case SMBC_MSG_CLOSEDIR:
	closedir_info = msg;
	//printf("smbc_message_loop: CLOSEDIR %d\n", closedir_info->fd);
	ack_info->value.iValue = smbc_closedir(closedir_info->fd);
	size = sizeof(int);
	break;
    case SMBC_MSG_READDIR:
	readdir_info = msg;
	//printf("smbc_message_loop: READDIR %d\n", readdir_info->fd);
	ack_info->value.pvValue = smbc_readdir(readdir_info->fd);
	size = sizeof(void *);
	break;
    case SMBC_MSG_TELLDIR:
	telldir_info = msg;
	//printf("smbc_message_loop: TELLDIR %d\n", telldir_info->dh);
	ack_info->value.llValue = smbc_telldir(telldir_info->dh);
	size = sizeof(long long);
	break;
    case SMBC_MSG_LSEEKDIR:
	lseekdir_info = msg;
	//printf("smbc_message_loop: LSEEKDIR %d\n", lseekdir_info->fd);
	ack_info->value.iValue = smbc_lseekdir(lseekdir_info->fd, lseekdir_info->offset);
	size = sizeof(int);
	break;
    case SMBC_MSG_MKDIR:
	mkdir_info = msg;
	//printf("smbc_message_loop: MKDIR %s\n", mkdir_info->fname);
	ack_info->value.iValue = smbc_mkdir(mkdir_info->fname, mkdir_info->mode);
	size = sizeof(int);
	break;
    case SMBC_MSG_RMDIR:
	rmdir_info = msg;
	//printf("smbc_message_loop: RMDIR %s\n", rmdir_info->durl);
	ack_info->value.iValue = smbc_rmdir(rmdir_info->durl);
	size = sizeof(int);
	break;
    case SMBC_MSG_STAT:
	stat_info = msg;
	//printf("smbc_message_loop: STAT %s\n", stat_info->fname);
	ack_info->value.iValue = smbc_stat(stat_info->fname, stat_info->st);
	size = sizeof(int);
	break;
    case SMBC_MSG_FSTAT:
	fstat_info = msg;
	//printf("smbc_message_loop: FSTAT %d\n", fstat_info->fd);
	ack_info->value.iValue = smbc_fstat(fstat_info->fd, fstat_info->st);
	size = sizeof(int);
	break;
#if 0
    case SMBC_MSG_CHOWN:
	chown_info = msg;
	//printf("smbc_message_loop: CHOWN %s\n", chown_info->url);
	ack_info->value.iValue = smbc_chown(chown_info->url, chown_info->owner, chown_info->group);
	size = sizeof(int);
	break;
    case SMBC_MSG_CHMOD:
	chmod_info = msg;
	//printf("smbc_message_loop: CHMOD %s\n", chmod_info->url);
	ack_info->value.iValue = smbc_chmod(chmod_info->url, chmod_info->mode);
	size = sizeof(int);
	break;
#endif
    case SMBC_MSG_REMOVEUNUSEDSERVERS:
	rus_info = msg;
	//printf("smbc_message_loop: REMOVEUNUSEDSERVERS %d\n", rus_info->msec);
	smbc_remove_unused_servers(rus_info->msec);
	size = 0;
	break;
    case SMBC_MSG_REINIT:
	//printf("smbc_message_loop: REINIT\n");
	ack_info->value.iValue = smbc_reinit();
	ack_info->value.iValue = nmbc_reinit() ? -1 : ack_info->value.iValue;
	size = sizeof(int);
	break;
    case NMBC_MSG_NAMEQUERY:
        {
            struct in_addr *ip_list;
            int count = 0, flags = 0;
            nq_info = msg;
            
            ip_list = nmbc_name_query(nq_info->name, 0, TRUE, FALSE,
                                      nq_info->bcast_addr, &count, &flags,
                                      NULL);
            if(!ip_list)
                ack_info->value.iValue = -1;
            else {
                /* only use 1st returned IP (?) */
                memcpy(nq_info->ip_addr, &ip_list[0], sizeof(struct in_addr));
                ack_info->value.iValue = 0;
            }
        }
        size = sizeof(int);
        break;
    default:
	printf("smbc_message_loop: Unknown message type %d\n", msg->type);
	ack_info->value.iValue = -1;
	size = sizeof(int);
    }
	ack_info->_errno=errno;
    ack_info->hdr.size = sizeof(SMBC_MSG) + sizeof(int) + size;
    ack_info->hdr.type = SMBC_MSG_ACK;
    return 1;
}

int smbc_message_loop(int fd)
{
    char buf[256];
    SMBC_MSG *msg;

    chkstack_init_thread("smbc_message_loop", 65536-260);

    printf("smbc_message_loop: Connected\n");
#if 0
    if (smbc_init(smbc_auth_fn, 0) < 0) {
	printf("smbc_message_loop: smb init failed, error %d\n", errno);
	close(fd);
	return 0;
    }
#endif

    //checksum_smbclient();

    msg = buf;
    do {

	//printf("smbc_message_loop: Waiting for message\n");

	if (!smbc_recv_msg(fd, msg, sizeof(buf)))
	    break;
	//printf("smbc_message_loop: Got message %d\n", msg->type);
	//sleep(1);

	if (!smbc_dispatch_msg(msg, msg))
	    break;

	//checksum_smbclient();

	//printf("smbc_message_loop: Response %d\n", ack_info->value.iValue);
	if (!smbc_send_msg(fd, msg))
	    break;

    } while (1);

    printf("smbc_message_loop: exiting\n");
    close(fd);
    return 0;
}

int smbc_send_msg(int fd, SMBC_MSG *msg)
{
    int i, j;

    //printf("smbc_send_msg: sending %d bytes\n", msg->size);
    for (i = 0; i < msg->size; i += j) {

	j = send(fd, ((char *)msg)+i, msg->size - i, 0);
	if (j < 0) {
	    printf("smbc_send_msg: Failed to send %d, error %d\n", j, errno);
	    return 0;
	}
    }
    //printf("smbc_send_msg: %d bytes sent\n", i);
    return i;
}

int smbc_recv_msg(int fd, SMBC_MSG *msg, int size)
{
    int i, j;

    for (i = 0; i < sizeof(SMBC_MSG); i += j) {

	j = recv(fd, ((char *)msg) + i, sizeof(SMBC_MSG) - i, 0);
	if (j <= 0) {
	    printf("smbc_recv_msg: Failed to receive %d, error %d\n", j, errno);
	    return 0;
	}
    }

    if (msg->size > size) {
	printf("smbc_recv_msg: Message too large %d\n", msg->size);
	return 0;
    }

    for (i; i < msg->size; i += j) {

	j = recv(fd, ((char *)msg) + i, msg->size - i, 0);
	if (j <= 0) {
	    printf("smbc_recv_msg: Failed to receive %d, error %d\n", j, errno);
	    return 0;
	}
    }
    return i;
}

pthread_t smbc_create_thread(void *(*thread_proc)(void *pv), void *pv)
{
    pthread_t id = 0;
    pthread_attr_t attr;

    pthread_attr_init(&attr);
    pthread_attr_setscope(&attr, 0);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    pthread_attr_setstacksize(&attr, 65536);
    pthread_create(&id, &attr, thread_proc, pv);
    pthread_attr_destroy(&attr);
    return id;
}

int smbc_listen_thread(void)
{
    int s, a, addrlen;
    struct sockaddr_in addr;

    s = socket(AF_INET, SOCK_STREAM, 0);
    if (s < 0) {
	printf("smbc_listen_thread: Couldn't create socket (error %d)\n", errno);
	return 0;
    }

    memset(&addr, 0, sizeof(addr));
    addr.sin_family      = AF_INET;
    addr.sin_port        = htons(SMBC_REMOTE_PORT);
    addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
    if (bind(s, &addr, sizeof(addr)) < 0) {
	printf("smbc_listen_thread: Couldn't bind to port (error %d)\n", errno);
	return 0;
    }

    if (listen(s, 5) < 0) {
	printf("smbc_listen_thread: Couldn't listen (error %d)\n", errno);
	return 0;
    }

    printf("smbc_listen_thread: Started\n");
    do {

	addrlen = sizeof(addr);
	a = accept(s, &addr, &addrlen);
	if (a >= 0) {
	    smbc_create_thread(smbc_message_loop, a);
	    //smbc_message_loop(a);
	    //close(a);
	}
    } while (1);
    return 0;
}

#define NOP()   	__asm("andeq\tr0,r0,r0")
#define NOP10() 	{ NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); }
#define NOP100()	{ NOP10(); NOP10(); NOP10(); NOP10(); NOP10(); NOP10(); NOP10(); NOP10(); NOP10(); NOP10(); }
#define NOP1000()       { NOP100(); NOP100(); NOP100(); NOP100(); NOP100(); NOP100(); NOP100(); NOP100(); NOP100(); NOP100(); }
#define NOP10000()	{ NOP1000(); NOP1000(); NOP1000(); NOP1000(); NOP1000(); NOP1000(); NOP1000(); NOP1000(); NOP1000(); NOP1000(); }

//char filldata[9995] = {-1};

int main(int argc, char *argv[])
{
    if (smbc_init(smbc_auth_fn, 0) < 0) {
	printf("smbclient: smb init failed, error %d\n", errno);
	return 1;
    }

    //smbc_create_thread(smbc_redir_thread, NULL);
    //checksum_smbclient();
    //malloc(1013);
    return smbc_listen_thread();
}
