
#define SMBC_MAX_PATH		(512)
#define SMBC_REMOTE_PORT	(666)

typedef struct _SMBC_MSG {
    int		size;
    int		type;
} SMBC_MSG;

typedef struct _SMBC_ACK_INFO {
    SMBC_MSG		hdr;
	int			_errno;	//Q: returning errno to libsmbremote
    union {
	int       	iValue;
	long	  	lValue;
	long long 	llValue;
	void *	        pvValue;
	char      	string[1];
	unsigned char 	buffer[1];
    } value;
} SMBC_ACK_INFO;
#define SMBC_MSG_ACK		(1)

typedef struct _SMBC_OPEN_INFO {
    SMBC_MSG	hdr;
    char *	fname;
    int		flags;
    int		mode;
} SMBC_OPEN_INFO;
#define SMBC_MSG_OPEN		(2)

typedef struct _SMBC_READ_INFO {
    SMBC_MSG	hdr;
    int         fd;
    int         count;
    void * 	buf;
} SMBC_READ_INFO;
#define SMBC_MSG_READ		(3)

typedef struct _SMBC_WRITE_INFO {
    SMBC_MSG	hdr;
    int         fd;
    int         count;
    void *      buf;
} SMBC_WRITE_INFO;
#define SMBC_MSG_WRITE		(4)

typedef struct _SMBC_LSEEK_INFO {
    SMBC_MSG	hdr;
    int         fd;
    long long   offset;
    int         whence;
} SMBC_LSEEK_INFO;
#define SMBC_MSG_LSEEK		(5)

typedef struct _SMBC_CLOSE_INFO {
    SMBC_MSG	hdr;
    int         fd;
} SMBC_CLOSE_INFO;
#define SMBC_MSG_CLOSE		(6)

typedef struct _SMBC_UNLINK_INFO {
    SMBC_MSG	hdr;
    char *      furl;
} SMBC_UNLINK_INFO;
#define SMBC_MSG_UNLINK		(7)

typedef struct _SMBC_LOCK_INFO {
    SMBC_MSG	hdr;
    int		fd;
    long long	offset;
    long	len;
    int         timeout;
    int         lock_type;
} SMBC_LOCK_INFO;
#define SMBC_MSG_LOCK		(8)

typedef struct _SMBC_UNLOCK_INFO {
    SMBC_MSG	hdr;
    int		fd;
    long long	offset;
    long	len;
} SMBC_UNLOCK_INFO;
#define SMBC_MSG_UNLOCK		(9)

typedef struct _SMBC_RENAME_INFO {
    SMBC_MSG	hdr;
    char *      ourl;
    char *      nurl;
} SMBC_RENAME_INFO;
#define SMBC_MSG_RENAME		(10)

typedef struct _SMBC_OPENDIR_INFO {
    SMBC_MSG	hdr;
    char *      fname;
} SMBC_OPENDIR_INFO;
#define SMBC_MSG_OPENDIR       	(11)

typedef struct _SMBC_CLOSEDIR_INFO {
    SMBC_MSG	hdr;
    int		fd;
} SMBC_CLOSEDIR_INFO;
#define SMBC_MSG_CLOSEDIR       (12)

typedef struct _SMBC_READDIR_INFO {
    SMBC_MSG	hdr;
    int		fd;
} SMBC_READDIR_INFO;
#define SMBC_MSG_READDIR     	(13)

typedef struct _SMBC_TELLDIR_INFO {
    SMBC_MSG	hdr;
    int		dh;
} SMBC_TELLDIR_INFO;
#define SMBC_MSG_TELLDIR        (14)

typedef struct _SMBC_LSEEKDIR_INFO {
    SMBC_MSG	hdr;
    int		fd;
    long long   offset;
} SMBC_LSEEKDIR_INFO;
#define SMBC_MSG_LSEEKDIR       (15)

typedef struct _SMBC_MKDIR_INFO {
    SMBC_MSG	hdr;
    char *      fname;
    int         mode;
} SMBC_MKDIR_INFO;
#define SMBC_MSG_MKDIR		(16)

typedef struct _SMBC_RMDIR_INFO {
    SMBC_MSG	hdr;
    char *      durl;
} SMBC_RMDIR_INFO;
#define SMBC_MSG_RMDIR		(17)

typedef struct _SMBC_STAT_INFO {
    SMBC_MSG	 hdr;
    char *       fname;
    struct stat *st;
} SMBC_STAT_INFO;
#define SMBC_MSG_STAT		(18)

typedef struct _SMBC_FSTAT_INFO {
    SMBC_MSG	 hdr;
    int          fd;
    struct stat *st;
} SMBC_FSTAT_INFO;
#define SMBC_MSG_FSTAT		(19)

typedef struct _SMBC_CHOWN_INFO {
    SMBC_MSG	hdr;
    char *      url;
    uid_t       owner;
    gid_t	group;
} SMBC_CHOWN_INFO;
#define SMBC_MSG_CHOWN          (20)

typedef struct _SMBC_CHMOD_INFO {
    SMBC_MSG	hdr;
    char *      url;
    mode_t      mode;
} SMBC_CHMOD_INFO;
#define SMBC_MSG_CHMOD          (21)

typedef struct _SMBC_REMOVEUNUSEDSERVERS_INFO {
    SMBC_MSG	hdr;
    long	msec;
} SMBC_REMOVEUNUSEDSERVERS_INFO;
#define SMBC_MSG_REMOVEUNUSEDSERVERS	(22)

#define SMBC_MSG_REINIT		(23)

typedef struct _NMBC_NAMEQUERY_INFO {
    SMBC_MSG hdr;
    const char *name;
    struct in_addr bcast_addr;
    struct in_addr *ip_addr;
} NMBC_NAMEQUERY_INFO;
#define NMBC_MSG_NAMEQUERY  (24)

