
#include <stdio.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "tiffio.h"

static tsize_t
_tiffReadProc(thandle_t fd, tdata_t buf, tsize_t size)
{
	return ((tsize_t) read((int) fd, buf, (size_t) size));
}

static tsize_t
_tiffWriteProc(thandle_t fd, tdata_t buf, tsize_t size)
{
	return ((tsize_t) write((int) fd, buf, (size_t) size));
}

static toff_t
_tiffSeekProc(thandle_t fd, toff_t off, int whence)
{
	return ((toff_t) lseek((int) fd, (off_t) off, whence));
}

static int
_tiffCloseProc(thandle_t fd)
{
	return (close((int) fd));
}


static toff_t
_tiffSizeProc(thandle_t fd)
{
	long fsize;
	return ((fsize = lseek((int) fd, 0, SEEK_END)) < 0 ? 0 : fsize);
}

int main(int argc, char *argv[])
{
    TIFF *tif;
    int fd;
    unsigned char *buf;
    unsigned long width, height, i;

    printf("tiff test\n");
    if (argc < 2)
	return 1;

#if 1
    printf("opening %s\n", argv[1]);
    fd = open(argv[1], O_RDONLY, 0);
    if (fd < 0) {
	printf("Couldn't open %s\n", argv[1]);
	return 1;
    }
#endif
    
    printf("Opening tif\n");
    //tif = TIFFOpen(argv[1], "r");
#if 1
    tif = TIFFClientOpen(argv[1], "r", (thandle_t)fd,
			 _tiffReadProc, _tiffWriteProc, _tiffSeekProc,
			 _tiffCloseProc, _tiffSizeProc, NULL, NULL);
#endif

    if (!tif) {
	printf("Couldn't open tiff\n");
	return 1;
    }
    printf("Opened tiff\n");
    TIFFSetFileno(tif, fd);

    TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &height);
    TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &width);
    printf("width %d height %d\n", width, height);

    buf = (unsigned char *)malloc(width * 4 + 4);

    for (i = 0; i < height; i++)
	TIFFReadScanline(tif, buf, i, 0);

    printf("Read %d scanlines\n", i);

    TIFFClose(tif);
    close(fd);
    return 0;
}


